use std::collections::{btree_map::Entry, BTreeMap};
use std::path::Path;

use anyhow::Context;
use itertools::Itertools;
use rayon::prelude::{IntoParallelIterator, IntoParallelRefIterator, ParallelIterator};
use ron::ser::PrettyConfig;

use crate::global_tables::objects::GOBJECTS;
use crate::sdk::repr::*;
use crate::util::DedupIter;
use crate::v2_types::{
    any_type::{self, AnyField, AnyObject, AnyProperty, AnyStruct},
    traits::{
        AsUObject, UArrayPropertyTrait, UBoolPropertyTrait, UEnumPropertyTrait, UEnumTrait,
        UObjectNonConst, UObjectTrait, UPropertyTrait, UStructPropertyTrait, *,
    },
    UClass, UEnum, UFunction, UProperty,
};
use crate::v2_types::{UScriptStruct, UStruct};

use super::repr::{
    Class, ClassField, ClassMethod, Enum, Package, ProcessedPackage, Type, UnrealType,
};
use super::FoldIntoPackages;

impl Sdk {
    pub fn build() -> anyhow::Result<Self> {
        let packages = Self::get_packages()?;

        let packages = packages
            .into_par_iter()
            .map(|pkg| {
                let package_ref = pkg
                    .package_object
                    .as_package_ref()
                    .context("object was not a package object.");

                let output =
                    package_ref.and_then(|package_ref| pkg.process().map(|pkg| (package_ref, pkg)));

                output
            })
            // remove empty packages
            .filter(|pkg| {
                pkg.as_ref()
                    .map(|(_, pkg)| !pkg.types.is_empty())
                    .unwrap_or(true)
            })
            .collect::<anyhow::Result<BTreeMap<_, _>>>()?;

        Ok(Self { packages })
    }

    pub fn to_json(&self) -> Result<String, serde_json::Error> {
        serde_json::to_string(self)
    }

    pub fn to_ron(&self) -> Result<String, ron::Error> {
        ron::ser::to_string_pretty(self, PrettyConfig::new().compact_arrays(true))
    }

    pub fn save_to_path<P: AsRef<Path>>(&self, path: P) -> Result<(), anyhow::Error> {
        self.to_ron()
            .context("failed to serialize")
            .and_then(|text| std::fs::write(path, text).context("failed to write to file."))
    }

    pub fn from_path_ron<P: AsRef<Path>>(path: P) -> anyhow::Result<Self> {
        let contents = std::fs::read(path)?;
        let new = ron::de::from_bytes::<Self>(&contents)?;

        Ok(new)
    }

    pub fn get_object(&self, key: &ObjectRef) -> Option<&UnrealType> {
        self.packages
            .get(&key.package)
            .and_then(|pkg| pkg.types.get(key))
    }

    fn get_packages() -> anyhow::Result<Vec<Package>> {
        let objects = GOBJECTS
            .read()
            .map_err(|_| anyhow::anyhow!("couldn't read gobject table"))?;
        let objects = objects.as_objects().context("no object array")?;

        let sorted_objects = objects
            .iter()
            .filter_map(|item| item.object())
            .fold_into_packages();

        let packages = sorted_objects
            .into_iter()
            .map(|(package_object, children)| Package {
                package_object,
                children,
            })
            .collect::<Vec<_>>();

        Ok(packages)
    }
}

fn default_or_anon(name: &str) -> bool {
    name.contains("Default__")
        || name.contains("<uninitialized")
        || name.contains("PLACEHOLDER-CLASS")
}

impl Enum {
    pub fn from_uenum(value: UEnum) -> anyhow::Result<UnrealType> {
        let values = value
            .names()
            .iter()
            .enumerate()
            .map(|(value, name)| {
                let name = name
                    .get_name()
                    .unwrap_or(format!("AnonymousVariant{value}"));
                (value, name)
            })
            .fold(BTreeMap::new(), |mut acc, (value, name)| {
                match acc.entry(name) {
                    Entry::Vacant(entry) => {
                        entry.insert(vec![value]);
                    }
                    Entry::Occupied(mut entry) => {
                        entry.get_mut().push(value);
                    }
                }
                acc
            });

        let values =
            values
                .into_iter()
                .fold(BTreeMap::<u32, String>::new(), |mut acc, (name, values)| {
                    if values.len() > 1 {
                        acc.extend(values.into_iter().map(|i| (i as u32, format!("{name}{i}"))));
                    } else {
                        acc.insert(values.into_iter().next().unwrap() as u32, name);
                    }
                    acc
                });

        Ok(UnrealType::Enum(Enum {
            obj_ref: value.as_uobject().object_ref(),
            name: value.get_name().context("could not get name")?,
            full_name: value.get_full_name().context("could not get full name")?,
            values,
        }))
    }
}

impl Class {
    pub fn from_uclass(class: UClass) -> anyhow::Result<UnrealType> {
        Ok(UnrealType::Class(Self {
            kind: StructKind::Object,
            ..Self::from_struct_inner(unsafe { class.cast() })?
        }))
    }

    pub fn from_actor(class: UClass) -> anyhow::Result<UnrealType> {
        Ok(UnrealType::Actor(Self {
            kind: StructKind::Actor,
            ..Self::from_struct_inner(unsafe { class.cast() })?
        }))
    }

    pub fn from_struct(strct: UScriptStruct) -> anyhow::Result<UnrealType> {
        Ok(UnrealType::Struct(Self {
            kind: StructKind::Struct,
            ..Self::from_struct_inner(unsafe { strct.cast() })?
        }))
    }

    fn from_struct_inner(strct: UStruct) -> anyhow::Result<Self> {
        let name = strct.get_name().context("could not get function name")?;
        let full_name = strct
            .get_full_name()
            .context("could not get full function name")?;

        let super_class = strct.super_field().and_then(|supr| {
            if supr != strct {
                Some(supr.as_uobject().object_ref())
            } else {
                None
            }
        });

        let (fields, methods) = Self::process_children(&strct);

        Ok(Self {
            obj_ref: strct.as_uobject().object_ref(),
            kind: StructKind::Struct,
            size: *strct.property_size() as u32,
            name,
            full_name,
            super_class,
            properties_size: *strct.property_size() as u32,
            min_alignment: *strct.min_alignment() as u32,
            fields,
            methods,
        })
    }

    /// returns tuple of (fields, methods)
    fn process_children(strct: &UStruct) -> (Vec<ClassField>, Vec<ClassMethod>) {
        enum ClassChild {
            Field(ClassField),
            Method(ClassMethod),
        }

        let (fields, methods) = strct
            .iter_children()
            .map(|field| -> anyhow::Result<Option<ClassChild>> {
                let child = match any_type::AnyField::from_field(field) {
                AnyField::Property(prop) => Some(ClassChild::Field(ClassField::from_uprop(prop)?)),
                 AnyField::Struct(strct)
                    if let AnyStruct::Function(func) = AnyStruct::from_struct(strct) => {
                        {
                            Some(ClassChild::Method(ClassMethod::from_ufunction(func)?))
                        }
                    }
                _ => None,
            };

                Ok(child)
            })
            .filter_map(|prop| match prop {
                Ok(field) => field,
                Err(err) => {
                    log::warn!("skipping field because: {err}");
                    None
                }
            })
            .fold(
                (Vec::new(), Vec::new()),
                |(mut fields, mut methods), child| {
                    match child {
                        ClassChild::Field(field) => fields.push(field),
                        ClassChild::Method(method) => methods.push(method),
                    }
                    (fields, methods)
                },
            );

        let fields = fields
            .into_iter()
            .dedup_with_by(
                |n, field| ClassField {
                    override_name: Some(format!("{}{}", field.name, n)),
                    ..field
                },
                |field: &ClassField| -> String { field.unique_name().to_string() },
            )
            .collect();

        let methods = methods
            .into_iter()
            .dedup_with_by(
                |n, field| ClassMethod {
                    override_name: Some(format!("{}{}", field.name, n)),
                    ..field
                },
                |field: &ClassMethod| -> String { field.unique_name().to_string() },
            )
            .collect();

        (fields, methods)
    }

    pub fn get_dependent_types(&self) -> Vec<ObjectRef> {
        self.fields
            .iter()
            .filter_map(|field| field.ty.dependent_type())
            .chain(
                self.methods
                    .iter()
                    .flat_map(|method| method.get_dependent_types()),
            )
            .chain(self.super_class)
            .collect::<Vec<_>>()
    }
}

impl ClassField {
    pub fn from_uprop(prop: UProperty) -> anyhow::Result<Self> {
        let name = prop
            .name()
            .get_name()
            .context("failed to retrieve field name")?;
        Ok(Self {
            offset: *prop.offset() as u32,
            size: *prop.element_size() as u32,
            flags: *prop.property_flags(),
            override_name: None,
            name,
            ty: resolve_type(prop).context("failed to get field type")?,
        })
    }

    pub fn get_dependent_types(&self) -> Vec<ObjectRef> {
        self.ty.dependent_type().into_iter().collect()
    }
}

impl ClassMethod {
    pub fn from_ufunction(func: UFunction) -> anyhow::Result<Self> {
        let name = func.get_name().context("could not get function name")?;
        let full_name = func
            .get_full_name()
            .context("could not get full function name")?;

        let params = func
            .iter_children()
            .filter_map(|field| match AnyField::from_field(field) {
                AnyField::Property(prop) => Some(prop),
                _ => None,
            })
            .map(|prop| ClassField::from_uprop(prop))
            .collect::<anyhow::Result<Vec<_>>>()?
            .into_iter()
            .dedup_with_by(
                |n, field| ClassField {
                    override_name: Some(format!("{}{}", field.name, n)),
                    ..field
                },
                |field: &ClassField| -> String { field.unique_name().to_string() },
            )
            .collect();

        Ok(ClassMethod {
            name,
            full_name,
            override_name: None,
            flags: *func.function_flags(),
            parameters: params,
        })
    }

    pub fn get_dependent_types(&self) -> Vec<ObjectRef> {
        self.parameters
            .iter()
            .filter_map(|field| field.ty.dependent_type())
            .collect()
    }
}

impl Package {
    pub fn process(self) -> anyhow::Result<ProcessedPackage> {
        let types = self
            .children
            .par_iter()
            // filter out anon or default objects
            .filter(|obj| {
                obj.get_name()
                    .map(|name| !default_or_anon(&name))
                    .unwrap_or(false)
            })
            .filter_map(|&object| -> Option<UnrealType> {
                let ty = match AnyObject::from_object(object) {
                    AnyObject::Field(field) => match AnyField::from_field(field) {
                        AnyField::Enum(my_enum) => Enum::from_uenum(my_enum),
                        AnyField::Struct(my_struct) => match AnyStruct::from_struct(my_struct) {
                            AnyStruct::Class(class) => Class::from_uclass(class),
                            AnyStruct::ScriptStruct(strct) => Class::from_struct(strct),
                            AnyStruct::Actor(actor) => Class::from_actor(actor),
                            _ => {
                                return None;
                            }
                        },
                        _ => {
                            return None;
                        }
                    },
                    _ => {
                        return None;
                    }
                };

                match ty {
                    Ok(ty) => {
                        return Some(ty);
                    }
                    Err(err) => {
                        log::error!("failed to serialize object {object:?}: {err}");
                        return None;
                    }
                }
            })
            .map(|ty| (ty.obj_ref(), ty))
            .collect::<BTreeMap<_, _>>();

        let dependencies = types
            .iter()
            .flat_map(|(_, ty)| ty.get_dependent_types().into_iter().map(|obj| obj.package))
            .sorted()
            .dedup()
            .collect::<Vec<_>>();

        Ok(ProcessedPackage {
            package_object: self
                .package_object
                .as_package_ref()
                .context("not actually a package object")?,
            types,
            dependencies,
            name: self
                .package_object
                .get_name()
                .context("failed to get package object name.")?,
            full_name: self
                .package_object
                .get_full_name()
                .context("failed to get package object name.")?,
        })
    }
}

fn resolve_type(prop: UProperty) -> anyhow::Result<Type> {
    let ty = match AnyProperty::from_prop(prop.clone()) {
        numeric @ AnyProperty::Numeric(_) => match numeric.as_any_numeric_property().unwrap() {
            any_type::AnyNumericProperty::U8(prop) => match prop.uenum() {
                Some(enm) => Type::Enum {
                    underlying: Box::new(Type::Primitive(PrimitiveType::U8)),
                    enum_type: enm.as_uobject().object_ref(),
                },
                None => Type::Primitive(PrimitiveType::U8),
            },
            any_type::AnyNumericProperty::U16(_) => Type::Primitive(PrimitiveType::U16),
            any_type::AnyNumericProperty::U32(_) => Type::Primitive(PrimitiveType::U32),
            any_type::AnyNumericProperty::U64(_) => Type::Primitive(PrimitiveType::U64),
            any_type::AnyNumericProperty::I8(_) => Type::Primitive(PrimitiveType::I8),
            any_type::AnyNumericProperty::I16(_) => Type::Primitive(PrimitiveType::I16),
            any_type::AnyNumericProperty::I32(_) => Type::Primitive(PrimitiveType::I32),
            any_type::AnyNumericProperty::I64(_) => Type::Primitive(PrimitiveType::I64),
            any_type::AnyNumericProperty::F32(_) => Type::Primitive(PrimitiveType::F32),
            any_type::AnyNumericProperty::F64(_) => Type::Primitive(PrimitiveType::F64),
            any_type::AnyNumericProperty::Other(_) => {
                return Err(anyhow::anyhow!("unhandled numeric property"));
            }
        },
        AnyProperty::Bool(prop) => Type::Primitive(PrimitiveType::Bool {
            byte_mask: *prop.byte_mask(),
            field_mask: *prop.field_mask(),
            byte_offset: *prop.byte_offset(),
            field_size: *prop.field_size(),
        }),
        AnyProperty::Interface(_) => return Err(anyhow::anyhow!("skipping interfaces for now")),
        object @ AnyProperty::ObjectBase(_) => {
            match object.as_any_object_base_property().unwrap() {
                any_type::AnyObjectBaseProperty::Object(obj) => Type::Class(
                    obj.property_class()
                        .context("object property missing properry class")?
                        .as_uobject()
                        .object_ref(),
                ),
                any_type::AnyObjectBaseProperty::WeakObject(obj) => Type::WeakPtr(
                    obj.property_class()
                        .context("weak ptr property missing property class.")?
                        .as_uobject()
                        .object_ref(),
                ),
                any_type::AnyObjectBaseProperty::LazyObject(obj) => Type::LazyPtr(
                    obj.property_class()
                        .context("lazy ptr property missing property class.")?
                        .as_uobject()
                        .object_ref(),
                ),
                any_type::AnyObjectBaseProperty::SoftObject(obj) => Type::SoftPtr(
                    obj.property_class()
                        .context("soft ptr property missing property class.")?
                        .as_uobject()
                        .object_ref(),
                ),
                asset @ any_type::AnyObjectBaseProperty::AssetObject(_) => {
                    match asset.as_any_asset_object_property().unwrap() {
                        any_type::AnyAssetObjectProperty::Class(class) => Type::AssetPtr(
                            class
                                .property_class()
                                .context("asset object property missing properry class")?
                                .as_uobject()
                                .object_ref(),
                        ),
                        any_type::AnyAssetObjectProperty::Object(_) => {
                            log::warn!("skipping AssetObject property.");
                            return Err(anyhow::anyhow!(
                                "unhandled asset object property (NOT AN ERROR)"
                            ));
                        }
                    }
                }
                any_type::AnyObjectBaseProperty::Other(_) => {
                    return Err(anyhow::anyhow!("unhandled object base property"));
                }
            }
        }
        AnyProperty::Array(array) => Type::Array(Box::new(resolve_type(
            array
                .inner()
                .context("array property inner type missing.")?,
        )?)),
        AnyProperty::Map(_) => unreachable!("not used in ARK"),
        AnyProperty::Str(_) => Type::String,
        AnyProperty::Text(_) => Type::Text,
        AnyProperty::Name(_) => Type::Name,
        // TODO: handle delegates, idk if they are ever useful though.
        AnyProperty::Delegate(_) => {
            log::warn!("skipping delegate property.");
            return Err(anyhow::anyhow!("skipping delegates for now"));
        }
        AnyProperty::MulticastDelegate(_) => {
            log::warn!("skipping delegate property.");
            return Err(anyhow::anyhow!("skipping multicast delegates for now"));
        }
        AnyProperty::Enum(enm) => Type::Enum {
            underlying: Box::new(resolve_type(
                enm.underlying_type()
                    .context("enum property was missing underlying type")?,
            )?),
            enum_type: enm
                .uenum()
                .context("enum property missing enum type")?
                .as_uobject()
                .object_ref(),
        },
        AnyProperty::Struct(class) => Type::Struct(
            class
                .ustruct()
                .context("struct property had no inner struct")?
                .as_uobject()
                .object_ref(),
        ),
        AnyProperty::Other(_) => {
            log::warn!("skipping unknown property type.");
            return Err(anyhow::anyhow!("unhandled property."));
        }
    };

    if *prop.array_dim() > 1 {
        Ok(Type::RawArray {
            ty: Box::new(ty),
            len: *prop.array_dim() as u32,
        })
    } else {
        Ok(ty)
    }
}
