use std::{collections::BTreeMap, fmt::Display};

use serde::{Deserialize, Serialize};

use crate::v2_types::{
    traits::{UObjectNonConst, UObjectTrait},
    EFunctionFlags, EPropertyFlags, UObject,
};

impl UObject {
    pub fn as_package_ref(&self) -> Option<PackageRef> {
        if self.is_package_object() {
            Some(PackageRef(*self.internal_index()))
        } else {
            None
        }
    }

    pub fn object_ref(&self) -> ObjectRef {
        ObjectRef {
            package: PackageRef(*self.package_object().internal_index()),
            object: *self.internal_index(),
        }
    }
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy, Serialize, Deserialize)]
pub struct PackageRef(u32);

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy, Serialize, Deserialize)]
pub struct ObjectRef {
    pub package: PackageRef,
    pub object: u32,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Sdk {
    pub packages: BTreeMap<PackageRef, ProcessedPackage>,
}

/// A package represents some group of packages that are related to another,
/// examples in ARK are `Engine`, `ShooterGame`, `SlateCore` or
/// `Buff_Companion_HLNA`.
#[derive(Debug)]
pub struct Package {
    /// the package object is the root object of the package.
    pub package_object: UObject,
    /// each package may contain other objects and classes which can be
    /// referenced by an `ObjectRef`, which is a unique identifier for each
    /// object.
    pub children: Vec<UObject>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct ProcessedPackage {
    /// the package object ref.
    pub package_object: PackageRef,
    pub name: String,
    pub full_name: String,
    /// all types extracted from this package referenced by their `ObjectRef`.
    pub types: BTreeMap<ObjectRef, UnrealType>,
    /// All other packages that types in this package depend on directly.
    pub dependencies: Vec<PackageRef>,
}

#[derive(Debug, Serialize, Deserialize)]
pub enum UnrealType {
    Class(Class),
    Struct(Class),
    Actor(Class),
    Enum(Enum),
}

impl UnrealType {
    pub fn super_class(&self) -> Option<ObjectRef> {
        match self {
            UnrealType::Class(class) | UnrealType::Struct(class) | UnrealType::Actor(class) => {
                class.super_class
            }
            UnrealType::Enum(_) => None,
        }
    }

    pub fn unique_name(&self) -> &str {
        // FIXME: change this to the unique name field in the future
        match self {
            UnrealType::Class(class) | UnrealType::Struct(class) | UnrealType::Actor(class) => {
                &class.name
            }
            UnrealType::Enum(enm) => &enm.name,
        }
    }

    pub fn obj_ref(&self) -> ObjectRef {
        match self {
            UnrealType::Class(obj) => obj.obj_ref,
            UnrealType::Struct(obj) => obj.obj_ref,
            UnrealType::Actor(obj) => obj.obj_ref,
            UnrealType::Enum(obj) => obj.obj_ref,
        }
    }

    pub fn get_dependent_types(&self) -> Vec<ObjectRef> {
        match self {
            UnrealType::Class(obj) | UnrealType::Struct(obj) | UnrealType::Actor(obj) => {
                obj.get_dependent_types()
            }
            UnrealType::Enum(_) => vec![],
        }
    }
}

#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
pub enum StructKind {
    Object,
    Actor,
    Struct,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Enum {
    pub obj_ref: ObjectRef,
    pub name: String,
    pub full_name: String,
    pub values: BTreeMap<u32, String>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Class {
    pub obj_ref: ObjectRef,
    pub kind: StructKind,
    pub size: u32,
    pub name: String,
    pub full_name: String,
    pub super_class: Option<ObjectRef>,
    pub properties_size: u32,
    pub min_alignment: u32,
    pub fields: Vec<ClassField>,
    pub methods: Vec<ClassMethod>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct ClassField {
    pub offset: u32,
    pub size: u32,
    pub name: String,
    pub override_name: Option<String>,
    pub flags: EPropertyFlags,
    pub ty: Type,
}

impl ClassField {
    pub fn is_return_param(&self) -> bool {
        self.flags.contains(EPropertyFlags::ReturnParm)
    }

    pub fn is_param(&self) -> bool {
        self.flags.contains(EPropertyFlags::Parm)
    }

    pub fn is_const_param(&self) -> bool {
        self.flags.contains(EPropertyFlags::ConstParm)
    }

    pub fn is_out_param(&self) -> bool {
        !self.is_const_param() && self.flags.contains(EPropertyFlags::OutParm)
    }

    pub fn is_rep(&self) -> bool {
        !self.flags.contains(EPropertyFlags::RepSkip)
    }

    pub fn unique_name(&self) -> &str {
        self.override_name
            .as_ref()
            .map(|s| s.as_str())
            .unwrap_or(self.name.as_str())
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct ClassMethod {
    pub name: String,
    pub override_name: Option<String>,
    pub full_name: String,
    pub flags: EFunctionFlags,
    pub parameters: Vec<ClassField>,
}

impl ClassMethod {
    pub fn is_static(&self) -> bool {
        self.flags.contains(EFunctionFlags::Static)
    }

    pub fn is_native(&self) -> bool {
        self.flags.contains(EFunctionFlags::Native)
    }

    /// this function is replicated to the server, might be exploitable.
    pub fn is_net_server(&self) -> bool {
        self.flags.contains(EFunctionFlags::NetServer)
    }

    pub fn unique_name(&self) -> &str {
        self.override_name
            .as_ref()
            .map(|s| s.as_str())
            .unwrap_or(self.name.as_str())
    }

    pub fn out_params(&self) -> Vec<&ClassField> {
        self.parameters
            .iter()
            .filter(|&param| param.is_out_param() || param.is_return_param())
            .collect::<Vec<_>>()
    }

    pub fn in_params(&self) -> Vec<&ClassField> {
        self.parameters
            .iter()
            .filter(|param| param.is_param() && !(param.is_out_param() || param.is_return_param()))
            .collect::<Vec<_>>()
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum PrimitiveType {
    Bool {
        byte_mask: u8,
        field_mask: u8,
        byte_offset: u8,
        field_size: u8,
    },
    U8,
    U16,
    U32,
    U64,
    I8,
    I16,
    I32,
    I64,
    F32,
    F64,
    Custom(String),
}

impl Display for PrimitiveType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PrimitiveType::Bool { .. } => f.write_str("bool"),
            PrimitiveType::U8 => f.write_str("u8"),
            PrimitiveType::U16 => f.write_str("u16"),
            PrimitiveType::U32 => f.write_str("u32"),
            PrimitiveType::U64 => f.write_str("u64"),
            PrimitiveType::I8 => f.write_str("i8"),
            PrimitiveType::I16 => f.write_str("i16"),
            PrimitiveType::I32 => f.write_str("i32"),
            PrimitiveType::I64 => f.write_str("i64"),
            PrimitiveType::F32 => f.write_str("f32"),
            PrimitiveType::F64 => f.write_str("f64"),
            PrimitiveType::Custom(custom) => f.write_str(&custom),
        }
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum Type {
    Ptr(Box<Type>),
    Ref(Box<Type>),
    WeakPtr(ObjectRef),
    SoftPtr(ObjectRef),
    LazyPtr(ObjectRef),
    AssetPtr(ObjectRef),
    Array(Box<Type>),
    Primitive(PrimitiveType),
    RawArray {
        ty: Box<Type>,
        len: u32,
    },
    Name,
    String,
    Text,
    Enum {
        underlying: Box<Type>,
        enum_type: ObjectRef,
    },
    Class(ObjectRef),
    Struct(ObjectRef),
}

impl Type {
    pub fn dependent_type(&self) -> Option<ObjectRef> {
        match self {
            Type::Ptr(ty) | Type::Ref(ty) | Type::Array(ty) | Type::RawArray { ty, .. } => {
                ty.dependent_type()
            }
            Type::WeakPtr(ty) | Type::SoftPtr(ty) | Type::LazyPtr(ty) | Type::AssetPtr(ty) => {
                Some(*ty)
            }
            Type::Enum { underlying, .. } => underlying.dependent_type(),
            Type::Class(ty) | Type::Struct(ty) => Some(*ty),
            _ => None,
        }
    }
}
